# Importation du Lexer et du Parser https://github.com/dabeaz/sly
from sly import Lexer
from sly import Parser


## PARTIE LEXER
class Lexer(Lexer):
    ignore = '\t '                                                      # Définition des caractères que l'on ignorera, ici on ignore les tabulations et les espaces

    # Ici on définit le nom des tokens
    tokens = { EGAL, PEGAL, SEGAL, MEGAL, SUP, INF,                     # Définition des opérateurs (==, !=, >=, <=, >, <)
               ASSIGNE, PLUS, MOINS, PUISSANCE, FOIS, DIVISE, MODULO,   # Définition des opérations (=, +, -, *, /, **, %)
               NOMBRE, CHAINE,                                          # Définition des types (float, string)
               SI,                                                      # Définition des conditions (if)
               NOM,                                                     # Définition d'une variable
               DEBOGUE                                                  # Définition du token debug, celui-ci permettra d'afficher toutes les variables en mémoires
    }
    
    
    # Puis on définit chaque token précisement, c'est à dire comment avec une expressions rationnelle le lexer peut le trouver
    ## OPERATEURS
    EGAL            = r'=='     # Voir documentation pour savoir pourquoi déclarer "==" avant "="
    PEGAL           = r'!='     # Pas égal = pégal
    SEGAL           = r'>='     # Supérieur ou égal = ségal
    MEGAL           = r'<='     # Moins ou égal = mégal
    SUP             = r'>'
    INF             = r'<'
    

    ## OPERATIONS
    ASSIGNE         = r'='
    PLUS            = r'\+'     # On oublie pas le \ qui permet d'échapper le caractère suivant (pareil pour la suite du programme)
    MOINS           = r'-'
    PUISSANCE       = r'\*\*'   # Voir documentation pour savoir pourquoi déclarer "**" avant "*" cf ligne au-dessus
    FOIS            = r'\*'
    DIVISE          = r'/'
    MODULO          = r'%'


    ## TYPES
    @_(r'\d+\.\d+|\d+')
    def NOMBRE(self, t):                    # Définition d'un nombre, \d+ signifie "un ou plusieurs chiffre" https://stackoverflow.com/a/40053690
        t.value = float(t.value)            # Convertit ce chiffre ou ce nombre en flottant (accepte également les entiers et les convertit en décimal)
        return t
    CHAINE =         r'\".*?\"'


    ## CONDITIONS
    SI =             r'SI'
    

    ## VARIABLES
    NOM =            r'[a-zA-Z_][a-zA-Z0-9_]*'   # Un nom de variable doit commencer par une lettre ou un underscore,
                                                 # suivit d'un caractère alphanumérique ou d'underscore

    ## DEBOGUE
    DEBOGUE =        r'&'                        # Pour faciliter le déboguage, il fallait un mot court et proche des doigts qui sont sur le clavier
                                                 # J'ai donc choisis '&', de plus il ne fallait pas qu'il soit utilisé autre part


    ## COMMENTAIRES
    @_(r'#.*')
    def COMMENT(self, t):
        pass


    # RETOUR CHARIOT (permet de compter les lignes, utile pour l'affichage d'erreur afin de faciliter le déboguage)
    @_(r'\n+')
    def newline(self, t):
        self.lineno += len(t.value)


    # ERREUR
    def error(self, t):
        print("Caractere inconnu : '%s'" % t.value[0])
        self.index += 1


## PARTIE PARSER
class Parser(Parser):
    # Initialisation des tokens du parser à l'aide des tokens du lexer
    tokens = Lexer.tokens


    # Définition de l'initialisation
    def __init__(self):
        self.env = { }


    # Déclaration d'une ligne vide
    @_('')
    def instruction(self, p):
        pass


    # Affichage de toutes les variables en mémoire
    @_('DEBOGUE')
    def instruction(self, p):
        return ('debogue', p[0])


    # Affichage du nombre entré, il suffit de le réécrire à l'écran
    @_('NOMBRE')
    def instruction(self, p):
        return ('nombre', p[0])


    # Affichage du chaîne de caractère entrée, il suffit de la réécrire à l'écran
    @_('CHAINE')
    def instruction(self, p):
        return ('affiche', p[0])


    # Déclaration de l'AFFICHAGE de la valeur d'une VARIABLE
    @_('NOM')
    def instruction(self, p):
        return ('variable', p[0])


    # Déclaration de l'assignation, l'addition, la soustraction, la multiplication, la division pour LES ENTIERS
    @_('NOM ASSIGNE NOMBRE',
       'NOM PLUS NOMBRE',
       'NOM MOINS NOMBRE',
       'NOM PUISSANCE NOMBRE',
       'NOM FOIS NOMBRE',
       'NOM DIVISE NOMBRE',
       'NOM MODULO NOMBRE')
    def instruction(self, p):
        return (p[1], p[0], p[2])
    

    # Déclaration de l'assignation d'une variable à l'aide une autre variable
    @_('NOM ASSIGNE NOM',
       'NOM PLUS NOM',
       'NOM MOINS NOM',
       'NOM PUISSANCE NOM',
       'NOM FOIS NOM',
       'NOM DIVISE NOM',
       'NOM MODULO NOM')
    def instruction(self, p):
        return (p[1], p[0], ('variable', p[2]))


    # Déclaration des opérateurs égal, pégal, ségal, mégal, sup, inf
    @_('NOM EGAL NOMBRE',
       'NOM PEGAL NOMBRE',
       'NOM SEGAL NOMBRE',
       'NOM MEGAL NOMBRE',
       'NOM SUP NOMBRE',
       'NOM INF NOMBRE',)
    def operateur(self, p):
        return (p[1], p[0], p[2])


    # Déclaration des opérateurs égal, pégal, ségal, mégal, sup, inf
    @_('NOM EGAL NOM',
       'NOM PEGAL NOM',
       'NOM SEGAL NOM',
       'NOM MEGAL NOM',
       'NOM SUP NOM',
       'NOM INF NOM',)
    def operateur(self, p):
        return (p[1], p[0], ('variable', p[2]))


    # Déclaration d'une condition
    @_('SI operateur instruction')
    def instruction(self, p):
        return ('si', p[1], p[2])



## PARTIE ÉXÉCUTION
class Execute:
    # Dans le cas de l'initialisation
    def __init__(self, arbre, env):
        self.env = env
        result = self.traverseLarbre(arbre)
        
        if result is not None:
            print(result)


    # Dans le cas d'un parcours de l'arbre
    def traverseLarbre(self, node):
        if node is None:                        # Si c'est vide on s'arrête là
            return None
        elif isinstance(node, float or str):    # Si c'est un flottant ou une chaine, on retourne le noeud
            return node

        # Sinon ça signifie qu'il faut un peu de logique pour arriver au fond du noeud :
        try:
            if node[0] == 'debogue':                ## Déboguage
                print(env)
            elif node[0] == 'nombre' or node[0] == 'affiche':  ## Si c'est un nombre ou une chaine
                return node[1]
            elif node[0] == 'si':                   ## Condition
                if self.traverseLarbre(node[1]) == 1:
                    return self.traverseLarbre(node[2])

            ## Opérations

            elif node[0] == 'variable':             ## Affichage
                return self.env[node[1]]
            elif node[0] == '=':                    ## Assignation
                self.env[node[1]] = self.traverseLarbre(node[2])
            elif node[0] == '+':                    ## Addition
                self.env[node[1]] = self.env[node[1]] + self.traverseLarbre(node[2])
            elif node[0] == '-':                    ## Soustraction
                self.env[node[1]] = self.env[node[1]] - self.traverseLarbre(node[2])
            elif node[0] == '*':                    ## Multiplication
                self.env[node[1]] = self.env[node[1]] * self.traverseLarbre(node[2])
            elif node[0] == '/':                    ## Division
                self.env[node[1]] = self.env[node[1]] / self.traverseLarbre(node[2])
            elif node[0] == '**':                   ## Puissance
                self.env[node[1]] = self.env[node[1]] ** self.traverseLarbre(node[2])
            elif node[0] == '%':                    ## Modulo
                self.env[node[1]] = self.env[node[1]] % self.traverseLarbre(node[2])

            ## Opérateurs

            elif node[0] == '==':                   ## Égal
                return self.env[node[1]] == self.traverseLarbre(node[2])
            elif node[0] == '!=':                   ## Pégal
                return self.env[node[1]] != self.traverseLarbre(node[2])
            elif node[0] == '>=':                   ## Ségal
                return self.env[node[1]] >= self.traverseLarbre(node[2])
            elif node[0] == '<=':                   ## Mégal
                return self.env[node[1]] <= self.traverseLarbre(node[2])
            elif node[0] == '>':                    ## Sup
               return self.env[node[1]] > self.traverseLarbre(node[2])
            elif node[0] == '<':                    ## Inf
                return self.env[node[1]] < self.traverseLarbre(node[2])
            return None

        except LookupError:                         ## Si il y a une exception, on écrit que l'on ne trouve pas la variable
            print("Variable '"+node[1]+"' non initialisee !")
            return 0                            ## Et on retourne 0


## Partie main
def main(debug = 0):
    lexer = Lexer()         # Initialisation du lexer
    parser = Parser()       # Initialisation du parser
    env = {}                # Initialisation de l'env (permet de stocker nos variables)

    # Si le programme n'est pas lancé dans l'IDE, on ne traite pas le fichier monProgramme.pk.
    if (debug == 1):
        ## Cette partie permet de lire le fichier monProgramme.pk afin de l'exécuter, c'est un peu plus pratique que de réécrire à chaque fois dans la console
        # L'extension pk étant une référence à l'enseignant qui m'a conduit à l'écriture de ces lignes. Je l'en remercie, car j'y ai beaucoup appris.
        monProgramme = open("monProgramme.pk", "r").readlines()
        for line in monProgramme:
            texte = line.strip()
            arbre = parser.parse(lexer.tokenize(texte))
            Execute(arbre, env)
        print("\n")

    ## Cette partie permet de lire l'entrée de la console afin de l'exécuter
    while True:
        texte = input('CISMUD > ')
        arbre = parser.parse(lexer.tokenize(texte))
        Execute(arbre, env)


# Permet de lancer le programme lorsque l'on passe par l'IDE Visual Studio Community
if __name__ == "__main__":
    main(1)
